@@ -25,6 +25,7 @@ struct kanshi_head {
 	struct wl_list link;
 
 	char *name, *description;
+	char *make, *model, *serial_number;
 	int32_t phys_width, phys_height; // mm
 	struct wl_list modes;
 
@@ -41,18 +42,29 @@ struct kanshi_head {
 
 struct kanshi_state {
 	bool running;
+	struct wl_display *display;
 	struct zwlr_output_manager_v1 *output_manager;
+#if KANSHI_HAS_VARLINK
+	struct VarlinkService *service;
+#endif
 
 	struct kanshi_config *config;
+	const char *config_arg;
 
 	struct wl_list heads;
 	uint32_t serial;
 	struct kanshi_profile *current_profile;
+	struct kanshi_profile *pending_profile;
 };
 
 struct kanshi_pending_profile {
+	uint32_t serial;
 	struct kanshi_state *state;
 	struct kanshi_profile *profile;
 };
 
+bool kanshi_reload_config(struct kanshi_state *state);
+
+int kanshi_main_loop(struct kanshi_state *state);
+
 #endif
