@@ -1,6 +1,7 @@
 #define _POSIX_C_SOURCE 200809L
 #include <assert.h>
 #include <errno.h>
+#include <fcntl.h>
 #include <getopt.h>
 #include <limits.h>
 #include <signal.h>
@@ -14,17 +15,27 @@
 #include "config.h"
 #include "kanshi.h"
 #include "parser.h"
+#include "ipc.h"
 #include "wlr-output-management-unstable-v1-client-protocol.h"
 
 #define HEADS_MAX 64
 
+static bool try_apply_profiles(struct kanshi_state *state);
+
 static bool match_profile_output(struct kanshi_profile_output *output,
 		struct kanshi_head *head) {
-	// TODO: improve vendor/model/serial matching
+	const char *make = head->make ? head->make : "Unknown";
+	const char *model = head->model ? head->model : "Unknown";
+	const char *serial_number =
+		head->serial_number ? head->serial_number : "Unknown";
+
+	char identifier[1024];
+	assert(sizeof(identifier) >= strlen(make) + strlen(model) + strlen(serial_number) + 3);
+	snprintf(identifier, sizeof(identifier), "%s %s %s", make, model, serial_number);
+
 	return strcmp(output->name, "*") == 0 ||
 		strcmp(output->name, head->name) == 0 ||
-		(strchr(output->name, ' ') != NULL &&
-		strstr(head->description, output->name) != NULL);
+		strcmp(output->name, identifier) == 0;
 }
 
 static bool match_profile(struct kanshi_state *state,
@@ -78,34 +89,46 @@ static struct kanshi_profile *match(struct kanshi_state *state,
 
 
 static void exec_command(char *cmd) {
-	pid_t pid, child;
+	pid_t child, grandchild;
 	// Fork process
-	if ((pid = fork()) == 0) {
-		// Fork child process again
+	if ((child = fork()) == 0) {
+		// Fork child process again so we can unparent the process
 		setsid();
 		sigset_t set;
 		sigemptyset(&set);
 		sigprocmask(SIG_SETMASK, &set, NULL);
-		if ((child = fork()) == 0) {
+
+		struct sigaction action;
+		sigfillset(&action.sa_mask);
+		action.sa_flags = 0;
+		action.sa_handler = SIG_DFL;
+		sigaction(SIGINT, &action, NULL);
+		sigaction(SIGQUIT, &action, NULL);
+		sigaction(SIGTERM, &action, NULL);
+		sigaction(SIGHUP, &action, NULL);
+
+		if ((grandchild = fork()) == 0) {
 			execl("/bin/sh", "/bin/sh", "-c", cmd, (void *)NULL);
-			fprintf(stderr, "Executing command '%s' failed: %s", cmd, strerror(errno));
-			exit(-1);
+			fprintf(stderr, "Executing command '%s' failed: %s\n", cmd, strerror(errno));
+			_exit(-1);
 		}
-		if (child < 0) {
+		if (grandchild < 0) {
 			fprintf(stderr, "Impossible to fork a new process to execute"
-					" command '%s': %s", cmd, strerror(errno));
-			exit(0);
+					" command '%s': %s\n", cmd, strerror(errno));
+			_exit(1);
 		}
-		exit(0); // Close child process
+		_exit(0); // Close child process
 	}
 
-	if (pid < 0) {
+	if (child < 0) {
 		perror("Impossible to fork a new process");
 		return;
 	}
 
 	// cleanup child process
-	waitpid(pid, NULL, 0);
+	if (waitpid(child, NULL, 0) < 0) {
+		perror("Impossible to clean up child process");
+	}
 }
 
 static void execute_profile_commands(struct kanshi_profile *profile) {
@@ -125,6 +148,9 @@ static void config_handle_succeeded(void *data,
 	fprintf(stderr, "configuration for profile '%s' applied\n",
 			pending->profile->name);
 	pending->state->current_profile = pending->profile;
+	if (pending->profile == pending->state->pending_profile) {
+		pending->state->pending_profile = NULL;
+	}
 	free(pending);
 }
 
@@ -134,6 +160,9 @@ static void config_handle_failed(void *data,
 	zwlr_output_configuration_v1_destroy(config);
 	fprintf(stderr, "failed to apply configuration for profile '%s'\n",
 			pending->profile->name);
+	if (pending->profile == pending->state->pending_profile) {
+		pending->state->pending_profile = NULL;
+	}
 	free(pending);
 }
 
@@ -144,6 +173,14 @@ static void config_handle_cancelled(void *data,
 	// Wait for new serial
 	fprintf(stderr, "configuration for profile '%s' cancelled, retrying\n",
 			pending->profile->name);
+	if (pending->profile == pending->state->pending_profile) {
+		pending->state->pending_profile = NULL;
+	}
+	if (pending->serial != pending->state->serial) {
+		// We've already received a new serial, try re-applying the profile
+		// immediately
+		try_apply_profiles(pending->state);
+	}
 	free(pending);
 }
 
@@ -153,15 +190,27 @@ static const struct zwlr_output_configuration_v1_listener config_listener = {
 	.cancelled = config_handle_cancelled,
 };
 
-static bool match_refresh(const struct kanshi_mode *mode, int refresh) {
+static bool match_refresh(const struct kanshi_mode *mode, int refresh, int *delta) {
 	int v = refresh - mode->refresh;
-	return abs(v) < 50;
+	int mode_delta = abs(v);
+	/* If we have a refresh, pick one with the lowest delta from our target.
+	 * Doing a simple fuzzy match that picks the greatest (due to ordering) here can lead us to picking a refresh
+	 * such as 120.01 or 60.01, which is problematic for two reasons:
+	 *  - Modes such as 4K 120.01Hz is too much for link bandwidth of DP 1.4 without DSC.
+	 *  - It becomes out of phase with the majority of content being displayed.
+	 */
+	if (mode_delta < 50 && mode_delta < *delta) {
+		*delta = mode_delta;
+		return true;
+	}
+	return false;
 }
 
 static struct kanshi_mode *match_mode(struct kanshi_head *head,
 		int width, int height, int refresh) {
 	struct kanshi_mode *mode;
 	struct kanshi_mode *last_match = NULL;
+	int mode_delta = INT32_MAX;
 
 	wl_list_for_each(mode, &head->modes, link) {
 		if (mode->width != width || mode->height != height) {
@@ -169,8 +218,8 @@ static struct kanshi_mode *match_mode(struct kanshi_head *head,
 		}
 
 		if (refresh) {
-			if (match_refresh(mode, refresh)) {
-				return mode;
+			if (match_refresh(mode, refresh, &mode_delta)) {
+				last_match = mode;
 			}
 		} else {
 			if (!last_match || mode->refresh > last_match->refresh) {
@@ -185,13 +234,17 @@ static struct kanshi_mode *match_mode(struct kanshi_head *head,
 static void apply_profile(struct kanshi_state *state,
 		struct kanshi_profile *profile,
 		struct kanshi_profile_output **matches) {
-	if (state->current_profile == profile) {
+	if (state->pending_profile == profile || state->current_profile == profile) {
 		return;
 	}
 
+	fprintf(stderr, "applying profile '%s'\n", profile->name);
+
 	struct kanshi_pending_profile *pending = calloc(1, sizeof(*pending));
+	pending->serial = state->serial;
 	pending->state = state;
 	pending->profile = profile;
+	state->pending_profile = profile;
 
 	struct zwlr_output_configuration_v1 *config =
 		zwlr_output_manager_v1_create_configuration(state->output_manager,
@@ -280,7 +333,11 @@ static void mode_handle_finished(void *data,
 		struct zwlr_output_mode_v1 *wlr_mode) {
 	struct kanshi_mode *mode = data;
 	wl_list_remove(&mode->link);
-	zwlr_output_mode_v1_destroy(mode->wlr_mode);
+	if (zwlr_output_mode_v1_get_version(mode->wlr_mode) >= 3) {
+		zwlr_output_mode_v1_release(mode->wlr_mode);
+	} else {
+		zwlr_output_mode_v1_destroy(mode->wlr_mode);
+	}
 	free(mode);
 }
 
@@ -370,12 +427,40 @@ static void head_handle_finished(void *data,
 		struct zwlr_output_head_v1 *wlr_head) {
 	struct kanshi_head *head = data;
 	wl_list_remove(&head->link);
-	zwlr_output_head_v1_destroy(head->wlr_head);
+	if (zwlr_output_head_v1_get_version(head->wlr_head) >= 3) {
+		zwlr_output_head_v1_release(head->wlr_head);
+	} else {
+		zwlr_output_head_v1_destroy(head->wlr_head);
+	}
 	free(head->name);
 	free(head->description);
+	free(head->make);
+	free(head->model);
+	free(head->serial_number);
 	free(head);
 }
 
+void head_handle_make(void *data,
+		struct zwlr_output_head_v1 *zwlr_output_head_v1,
+		const char *make) {
+	struct kanshi_head *head = data;
+	head->make = strdup(make);
+}
+
+void head_handle_model(void *data,
+		struct zwlr_output_head_v1 *zwlr_output_head_v1,
+		const char *model) {
+	struct kanshi_head *head = data;
+	head->model = strdup(model);
+}
+
+void head_handle_serial_number(void *data,
+		struct zwlr_output_head_v1 *zwlr_output_head_v1,
+		const char *serial_number) {
+	struct kanshi_head *head = data;
+	head->serial_number = strdup(serial_number);
+}
+
 static const struct zwlr_output_head_v1_listener head_listener = {
 	.name = head_handle_name,
 	.description = head_handle_description,
@@ -387,6 +472,9 @@ static const struct zwlr_output_head_v1_listener head_listener = {
 	.transform = head_handle_transform,
 	.scale = head_handle_scale,
 	.finished = head_handle_finished,
+	.make = head_handle_make,
+	.model = head_handle_model,
+	.serial_number = head_handle_serial_number,
 };
 
 static void output_manager_handle_head(void *data,
@@ -404,21 +492,25 @@ static void output_manager_handle_head(void *data,
 	zwlr_output_head_v1_add_listener(wlr_head, &head_listener, head);
 }
 
-static void output_manager_handle_done(void *data,
-		struct zwlr_output_manager_v1 *manager, uint32_t serial) {
-	struct kanshi_state *state = data;
-	state->serial = serial;
-
+static bool try_apply_profiles(struct kanshi_state *state) {
 	assert(wl_list_length(&state->heads) <= HEADS_MAX);
 	// matches[i] gives the kanshi_profile_output for the i-th head
 	struct kanshi_profile_output *matches[HEADS_MAX];
 	struct kanshi_profile *profile = match(state, matches);
 	if (profile != NULL) {
-		fprintf(stderr, "applying profile '%s'\n", profile->name);
 		apply_profile(state, profile, matches);
-	} else {
-		fprintf(stderr, "no profile matched\n");
+		return true;
 	}
+	fprintf(stderr, "no profile matched\n");
+	return false;
+}
+
+static void output_manager_handle_done(void *data,
+		struct zwlr_output_manager_v1 *manager, uint32_t serial) {
+	struct kanshi_state *state = data;
+	state->serial = serial;
+
+	try_apply_profiles(state);
 }
 
 static void output_manager_handle_finished(void *data,
@@ -437,8 +529,12 @@ static void registry_handle_global(void *data, struct wl_registry *registry,
 	struct kanshi_state *state = data;
 
 	if (strcmp(interface, zwlr_output_manager_v1_interface.name) == 0) {
+		uint32_t bind_version = 2;
+		if (version >= 3) {
+			bind_version = 3;
+		}
 		state->output_manager = wl_registry_bind(registry, name,
-			&zwlr_output_manager_v1_interface, 1);
+			&zwlr_output_manager_v1_interface, bind_version);
 		zwlr_output_manager_v1_add_listener(state->output_manager,
 			&output_manager_listener, state);
 	}
@@ -477,6 +573,40 @@ static struct kanshi_config *read_config(const char *config) {
 	return parse_config(config_path);
 }
 
+static void destroy_config(struct kanshi_config *config) {
+	struct kanshi_profile *profile, *tmp_profile;
+	wl_list_for_each_safe(profile, tmp_profile, &config->profiles, link) {
+		struct kanshi_profile_output *output, *tmp_output;
+		wl_list_for_each_safe(output, tmp_output, &profile->outputs, link) {
+			free(output->name);
+			wl_list_remove(&output->link);
+			free(output);
+		}
+		struct kanshi_profile_command *command, *tmp_command;
+		wl_list_for_each_safe(command, tmp_command, &profile->commands, link) {
+			free(command->command);
+			wl_list_remove(&command->link);
+			free(command);
+		}
+		wl_list_remove(&profile->link);
+		free(profile);
+	}
+	free(config);
+}
+
+bool kanshi_reload_config(struct kanshi_state *state) {
+	fprintf(stderr, "reloading config\n");
+	struct kanshi_config *config = read_config(state->config_arg);
+	if (config != NULL) {
+		destroy_config(state->config);
+		state->config = config;
+		state->pending_profile = NULL;
+		state->current_profile = NULL;
+		return try_apply_profiles(state);
+	}
+	return false;
+}
+
 static const char usage[] = "Usage: %s [options...]\n"
 "  -h, --help           Show help message and quit\n"
 "  -c, --config <path>  Path to config file.\n";
@@ -488,14 +618,13 @@ static const struct option long_options[] = {
 };
 
 int main(int argc, char *argv[]) {
-	char *config_arg = NULL;
+	const char *config_arg = NULL;
 
 	int opt;
 	while ((opt = getopt_long(argc, argv, "hc:", long_options, NULL)) != -1) {
 		switch (opt) {
 		case 'c':
-			free(config_arg);
-			config_arg = strdup(optarg);
+			config_arg = optarg;
 			break;
 		case 'h':
 			fprintf(stderr, usage, argv[0]);
@@ -511,8 +640,6 @@ int main(int argc, char *argv[]) {
 		return EXIT_FAILURE;
 	}
 
-	free(config_arg);
-
 	struct wl_display *display = wl_display_connect(NULL);
 	if (display == NULL) {
 		fprintf(stderr, "failed to connect to display\n");
@@ -521,8 +648,17 @@ int main(int argc, char *argv[]) {
 
 	struct kanshi_state state = {
 		.running = true,
+		.display = display,
 		.config = config,
+		.config_arg = config_arg,
 	};
+	int ret = EXIT_SUCCESS;
+#if KANSHI_HAS_VARLINK
+	if (kanshi_init_ipc(&state) != 0) {
+		ret = EXIT_FAILURE;
+		goto done;
+	}
+#endif
 	wl_list_init(&state.heads);
 
 	struct wl_registry *registry = wl_display_get_registry(display);
@@ -533,12 +669,17 @@ int main(int argc, char *argv[]) {
 	if (state.output_manager == NULL) {
 		fprintf(stderr, "compositor doesn't support "
 			"wlr-output-management-unstable-v1\n");
-		return EXIT_FAILURE;
+		ret = EXIT_FAILURE;
+		goto done;
 	}
 
-	while (state.running && wl_display_dispatch(display) != -1) {
-		// This space intentionally left blank
-	}
+	ret = kanshi_main_loop(&state);
 
-	return EXIT_SUCCESS;
+done:
+#if KANSHI_HAS_VARLINK
+	kanshi_free_ipc(&state);
+#endif
+	wl_display_disconnect(display);
+
+	return ret;
 }
